/**
 * Copyright (C) 2009-2016 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.docgen.collector.impl;

import java.util.ArrayList;
import java.util.Collection;

import org.apache.log4j.Logger;
import org.tanjakostic.jcleancim.docgen.collector.BookmarkRegistry;
import org.tanjakostic.jcleancim.docgen.collector.DocgenConfig;
import org.tanjakostic.jcleancim.docgen.collector.PropertiesDoc;
import org.tanjakostic.jcleancim.docgen.collector.TableSpec;
import org.tanjakostic.jcleancim.docgen.collector.WAX;
import org.tanjakostic.jcleancim.model.UmlAssociationEnd;
import org.tanjakostic.jcleancim.model.UmlAssociationEnd.AssociationEndPair;
import org.tanjakostic.jcleancim.model.UmlClass;

/**
 * Data required for documentation of IEC 61850 association ends of the class with the other end
 * class; applicable to meta-model classes only. For the layout, see {@link PropertiesDoc}.
 *
 * @author tatjana.kostic@ieee.org
 * @author laurent.guise@free.fr
 * @version $Id: CustomAssocEndsDoc.java 17 2016-07-23 20:38:15Z dev978 $
 */
class CustomAssocEndsDoc extends AbstractPropertiesDoc {
	private static final Logger _logger = Logger.getLogger(CustomAssocEndsDoc.class.getName());

	public static final String INTRO_FMT = " shows all relations of %s with other classes.";
	public static final String CAPTION_FMT = "Relations of %s with other classes";

	/**
	 * Constructor.
	 */
	public CustomAssocEndsDoc(DocgenConfig docgenCfg, UmlClass c, BookmarkRegistry bmRegistry) {
		super(docgenCfg, c, "ae", String.format(INTRO_FMT, c.getName()),
				String.format(CAPTION_FMT, c.getName()), TableSpec.CUSTOM_ASSOC_ENDS, null,
				bmRegistry);

		log(_logger, "---- collecting doc for assocEnds of " + c.getQualifiedName() + " ...");

		Collection<AssociationEndPair> retainedNatives = new ArrayList<AssociationEndPair>();
		Collection<AssociationEndPair> retainedInheriteds = new ArrayList<AssociationEndPair>();
		super.filterAssociationEnds(c, retainedNatives, retainedInheriteds);

		String bookmarkID = null; // we don't point from hyperlinks to any association end here

		for (AssociationEndPair aePair : retainedNatives) {
			UmlAssociationEnd otherEnd = aePair.getOtherEnd();
			UmlAssociationEnd myEnd = aePair.getMyEnd();
			String name = otherEnd.getName();
			String mult = otherEnd.getMultiplicity().getBounds();
			String type = prepareForHyperlink(otherEnd.getType());
			// String cond = get61850PresenceCondition(otherEnd);

			/*
			 * Currently, EA actually does not allow markup in the notes for association ends; if we
			 * use html returned by EA, the format is screwed (i.e., we don't get what we see).
			 * Therefore, we use raw text always. Commented line is for when EA starts supporting
			 * html format in association end notes.
			 */
			// CellText desc = deduceCellText(getDeprecatedTextAsPrefix(otherEnd), otherEnd);

			CellText desc = deduceCellText(getDeprecatedTextAsPrefix(otherEnd),
					otherEnd.getDescription(), null, otherEnd);

			EntryDocImpl entry = EntryDocImpl.createData(bookmarkID, desc.formatInfo, name, mult,
					type, desc.text);
			addEntry(entry);

			// for xml:
			initRawData(entry, otherEnd);
			initAssocEndRawData(entry, otherEnd, myEnd);
		}
		for (AssociationEndPair aePair : retainedInheriteds) {
			UmlAssociationEnd otherEnd = aePair.getOtherEnd();
			UmlAssociationEnd myEnd = aePair.getMyEnd();
			String name = otherEnd.getName();
			String mult = otherEnd.getMultiplicity().getBounds();
			String type = prepareForHyperlink(otherEnd.getType());
			// String cond = get61850PresenceCondition(otherEnd);
			String description = getDeprecatedTextAsPrefix(otherEnd) + INHERITED_FROM
					+ prepareForHyperlink(myEnd.getType());

			EntryDocImpl entry = EntryDocImpl.createData(bookmarkID, null, name, mult, type,
					description);
			addEntry(entry);

			// for xml:
			initRawData(entry, otherEnd, myEnd.getType().getName());
			initAssocEndRawData(entry, otherEnd, myEnd);
		}

		log(_logger, ">>>>>> " + toString());
	}

	private static void initAssocEndRawData(EntryDocImpl entry, UmlAssociationEnd otherEnd,
			UmlAssociationEnd myEnd) {
		entry.putCell(WAX.A_myMult, myEnd.getMultiplicity().getBounds());
		entry.putCell(WAX.A_type, otherEnd.getType().getName());
		entry.putCell(WAX.A_mult, otherEnd.getMultiplicity().getBounds());
		// entry.putCell(WAX.A_presCond, get61850PresenceCondition(otherEnd));
		// entry.addRawField(WAX.A_presCondArgs, "FIXME!");
	}

	// /** Returns presence condition for association ends. */
	// protected static final String get61850PresenceCondition(UmlAssociationEnd ae) {
	// return /* (ae.isConditional()) ? ae.getConditionNames() : */
	// ae.getMultiplicity().isOptional() ? UmlMultiplicity.Kind.O.toString()
	// : UmlMultiplicity.Kind.M.toString();
	// }
}
