/**
 * Copyright (C) 2009-2016 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.docgen.writer.word.doc;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.tanjakostic.jcleancim.docgen.UnsupportedInputFormatException;
import org.tanjakostic.jcleancim.docgen.UnsupportedOutputFormatException;
import org.tanjakostic.jcleancim.docgen.collector.ColumnSpec;
import org.tanjakostic.jcleancim.docgen.collector.EntryDoc;
import org.tanjakostic.jcleancim.docgen.collector.EntryDoc.Kind;
import org.tanjakostic.jcleancim.docgen.collector.PropertiesDoc;
import org.tanjakostic.jcleancim.docgen.writer.Caption;
import org.tanjakostic.jcleancim.docgen.writer.Caption.CaptionKind;
import org.tanjakostic.jcleancim.docgen.writer.Cursor;
import org.tanjakostic.jcleancim.docgen.writer.CursorList;
import org.tanjakostic.jcleancim.docgen.writer.Placeholder;
import org.tanjakostic.jcleancim.docgen.writer.Range;
import org.tanjakostic.jcleancim.docgen.writer.Style;
import org.tanjakostic.jcleancim.docgen.writer.word.AbstractWordWriter;
import org.tanjakostic.jcleancim.docgen.writer.word.WordPatternFinder;
import org.tanjakostic.jcleancim.docgen.writer.word.WordWriterInput;
import org.tanjakostic.jcleancim.util.ApplicationException;
import org.tanjakostic.jcleancim.util.Util;

import com.jacob.activeX.ActiveXComponent;
import com.jacob.com.Dispatch;
import com.jacob.com.EnumVariant;
import com.jacob.com.Variant;

/**
 * This was a huuuuuuuuuge pain, but satisfies the needs of {@link DocWordWriter} ! Most of it
 * TTD-ed.
 * <p>
 * Because jacob has no API for constants, when we have to provide constants as arguments to VBA
 * methods, we found the values for those constants with ObjectType browser in VBA (macro) editor.
 * <p>
 * In general, we split the processing into two steps:
 * <ol>
 * <li>scanning placeholders (and recording ranges), and,</li>
 * <li>overwriting placeholders with the content.</li>
 * </ol>
 * For the case where hyperlink creation is enabled, in overwriting user-defined placeholder (with
 * the content from e.g. UML packages), we actually insert / write internal hyperlink placeholders
 * that have information about the text we'd like to see and the bookmark we'd like to link it to.
 * Then we repeat the above two steps once more, when we know all the content within the document
 * and all the bookmarks available to link to.
 *
 * @author tatjana.kostic@ieee.org
 * @author laurent.guise@free.fr
 * @version $Id: DocWordWriter.java 15 2016-07-12 15:11:42Z dev978 $
 */
public class DocWordWriter extends AbstractWordWriter<Dispatch> {
	private static final Logger _logger = Logger.getLogger(DocWordWriter.class.getName());

	private Dispatch _wordApp;
	private Dispatch _wordDoc;

	private Dispatch _bookmarks;
	private Dispatch _hyperlinks;

	/**
	 * Constructs this instance and copies input file into output directory; this writer will write
	 * into that copy to produce the final document, by replacing placeholders found in it.
	 *
	 * @throws UnsupportedInputFormatException
	 *             if configured with template in unsupported format
	 * @throws UnsupportedOutputFormatException
	 *             if configured with output in unsupported format
	 * @throws IOException
	 *             if fails to copy input template into output directory.
	 */
	public DocWordWriter(WordWriterInput input) throws UnsupportedInputFormatException,
			UnsupportedOutputFormatException, IOException {
		super(input);
	}

	// ===============================

	private EnumVariant getIterator(Dispatch container, String collectionName) {
		Dispatch collection = Dispatch.get(container, collectionName).getDispatch();
		return new EnumVariant(collection);
	}

	/** Returns Word range. */
	private Dispatch getDocumentContent() {
		return Dispatch.get(_wordDoc, "Content").toDispatch();
	}

	/** Wraps <code>obj</code>'s "Range" property into our Range. */
	private Range<Dispatch> createRangeFromRangeOf(Dispatch obj) {
		return createRange(getRangeOf(obj));
	}

	private Dispatch getRangeOf(Dispatch obj) {
		return Dispatch.get(obj, "Range").toDispatch();
	}

	private void select(Dispatch obj) {
		Dispatch.call(obj, "Select");
	}

	// ---------------- Clipboard (Figures and markup text) -------------------

	private static void pasteFromClipboard(Dispatch range) throws ApplicationException {
		try {
			Dispatch.call(range, "Paste");
		} catch (Exception e) {
			e.printStackTrace();
			throw new ApplicationException("pasteFromClipboard failed", e);
		}
	}

	// private static final Variant wdPasteHTML = getVariant(10);
	// private static final Variant wdInLine = getVariant(0);
	private static final Variant wdFormatSurroundingFormattingWithEmphasis = DocWordHelper
			.getVariant(20);

	private static void pasteHtmlFromClipboard(Dispatch range) throws ApplicationException {
		try {
			Dispatch.call(range, "PasteAndFormat", wdFormatSurroundingFormattingWithEmphasis);
		} catch (Exception e) {
			e.printStackTrace();
			throw new ApplicationException("pasteHtmlFromClipboard failed", e);
		}
	}

	// -------------------------------------

	private void setRangeStyle(Dispatch range, Style style) {
		Dispatch.put(range, "Style", style.getName());
	}

	// =============================== Tables ===================================

	public static final boolean ORIG_CREATE_TABLE = true;

	private int doInsertTable(Range<Dispatch> range, PropertiesDoc doc, Style tabhead,
			boolean addBookmarks) {
		Level level = Level.DEBUG;

		long totStart = System.currentTimeMillis();
		_logger.info("--- insertTable() " + doc.getRowCount() + " rows: " + doc.getCaptionText());

		long start = totStart;
		appendRawTextInNewParagraphWithStyle(range, "", Style.tabcell);
		long pause = System.currentTimeMillis();
		_logger.log(level, "  ... " + (pause - start)
				+ " ms for appendRawTextInNewParagraphWithStyle()");

		Dispatch table = null;
		if (ORIG_CREATE_TABLE) {
			start = pause;
			table = createTable(range, doc.getColumnCount(), doc.getRowCount());
			pause = System.currentTimeMillis();
			_logger.log(level, "  ... " + (pause - start) + " ms for createTable()");

			start = pause;
			fillValues(table, _wordApp, doc, addBookmarks);
			pause = System.currentTimeMillis();
			_logger.log(level, "  ... " + (pause - start) + " ms for fillValues()");
		} else {
			start = pause;
			table = createTable(range, doc);
			pause = System.currentTimeMillis();
			_logger.log(level, "  ... " + (pause - start) + " ms for createTableWithValues()");
		}

		start = pause;
		formatTable(table, doc, tabhead, level); // format after filling: it may merge cells
		pause = System.currentTimeMillis();
		_logger.log(level, "  ... " + (pause - start) + " ms for formatTable()");

		int rangeEnd = createRangeFromRangeOf(table).getEnd();
		range.setEnd(rangeEnd);

		_logger.info("... " + (pause - totStart) + " ms total for insertTable()");
		return rangeEnd;
	}

	private static final Variant wdWord8TableBehavior = DocWordHelper.getVariant(0);
	private static final Variant wdAutoFitFixed = DocWordHelper.getVariant(0);

	private static Dispatch createTable(Range<Dispatch> range, int numColumns, int numRows) {
		Dispatch tables = Dispatch.get(range.getObject(), "Tables").toDispatch();
		return Dispatch.call(tables, "Add", range.getObject(), DocWordHelper.getVariant(numRows),
				DocWordHelper.getVariant(numColumns), wdWord8TableBehavior, wdAutoFitFixed)
				.getDispatch();
	}

	private static final Variant wdSeparateByCommas = DocWordHelper.getVariant(2);

	// FIXME: We have tried this (which is effectively somewhat faster), but found no way
	// to have Word accept e.g. new line (row) or comma (cell separator) which is included in
	// cell value. So, the split to cells is wrong...
	private static Dispatch createTable(Range<Dispatch> range, PropertiesDoc doc) {
		String cellSep = ",";
		Variant cellSepVb = wdSeparateByCommas;

		String csvTxt = asCsv(doc, cellSep);
		// System.out.println(">>>+++ " + csvTxt);

		range.setText(csvTxt);

		// ConvertToTable([Separator],
		// [NumRows], [NumColumns], [InitialColumnWidth], [Format],
		// [ApplyBorders], [ApplyShading], [ApplyFont], [ApplyColor],
		// [ApplyHeadingRows], [ApplyLastRow], [ApplyFirstColumn], [ApplyLastColumn], [AutoFit],
		// [AutoFitBehavior], [DefaultTableBehavior]) As Table

		return Dispatch.call(range.getObject(), "ConvertToTable", cellSepVb,
				DocWordHelper.getVariant(doc.getRowCount()),
				DocWordHelper.getVariant(doc.getColumnCount()), Variant.DEFAULT, Variant.DEFAULT,
				Variant.DEFAULT, Variant.DEFAULT, Variant.DEFAULT, Variant.DEFAULT,
				Variant.DEFAULT, Variant.DEFAULT, Variant.DEFAULT, Variant.DEFAULT,
				Variant.DEFAULT, wdAutoFitFixed, wdWord8TableBehavior).getDispatch();
	}

	public static String asCsv(PropertiesDoc doc, String cellSep) {
		StringBuilder sb = new StringBuilder();

		char lineSep = 13; // ANSI13 for line separator

		int rowCount = doc.getRowCount();
		int colCount = doc.getColumnCount();
		String[][] values = doc.getCellValues();

		for (int i = 0; i < rowCount; ++i) {
			for (int j = 0; j < colCount; ++j) {
				sb.append(StringEscapeUtils.escapeCsv(values[i][j]));
				if (j != colCount - 1) {
					sb.append(cellSep);
				}
			}
			if (i != rowCount - 1) {
				sb.append(lineSep);
			}
		}
		return sb.toString();
	}

	private Dispatch getSelection(Dispatch app) {
		return Dispatch.get(app, "Selection").getDispatch();
	}

	static enum FillCellKind {
		/** new default; the fastest */
		range,

		rangeSelection, selection,

		/** original; the slowest */
		cellIndexing;
	}

	// if changing this value, ensure to implement bookmarkID handling!
	static final FillCellKind FILL_CELL_METHOD = FillCellKind.range;

	private void fillValues(Dispatch table, Dispatch app, PropertiesDoc doc, boolean addBookmarks) {
		int rowCount = doc.getRowCount();
		int colCount = doc.getColumnCount();
		String[][] values = doc.getCellValues();
		String[] bookmarkIDs = doc.getBookmarkIDs();

		switch (FILL_CELL_METHOD) {
			case range: {
				// Dispatch tabRange = asRange(table).getObject();
				// Dispatch cells = Dispatch.get(tabRange, "Cells").getDispatch();
				EnumVariant enumVariant = getIterator(getRangeOf(table), "Cells");

				for (int i = 0; i < rowCount; ++i) {
					String bookmarkID = bookmarkIDs[i];
					Range<Dispatch> singleCell = null;
					for (int j = 0; j < colCount; ++j) {
						if (enumVariant.hasMoreElements()) {
							Dispatch cell = enumVariant.nextElement().toDispatch();
							String value = values[i][j];
							if (!value.isEmpty()) {
								Range<Dispatch> cellAsRange = createRangeFromRangeOf(cell);
								cellAsRange.setText(value);
								if (singleCell == null && addBookmarks && (bookmarkID != null)) {
									singleCell = cellAsRange;
									insertBookmark(singleCell, bookmarkID);
									getInput().getBookmarkRegistry().markAsAvailableInDocument(
											bookmarkID);
									_logger.trace("  DocWordWriter.fillValues(): added bookmark '"
											+ bookmarkID + "' to Word document and to used list.");
								}
							}
						}
					}
				}
				break;
			}
			case rangeSelection: {
				Dispatch tabRange = createRangeFromRangeOf(table).getObject();
				select(tabRange);
				Dispatch selection = getSelection(app);

				Dispatch cells = Dispatch.get(selection, "Cells").getDispatch();
				EnumVariant enumVariant = new EnumVariant(cells);

				for (int i = 0; i < rowCount; ++i) {
					for (int j = 0; j < colCount; ++j) {
						if (enumVariant.hasMoreElements()) {
							Dispatch cell = enumVariant.nextElement().toDispatch();
							String value = values[i][j];
							if (!value.isEmpty()) {
								createRangeFromRangeOf(cell).setText(value);
							}
						}
					}
				}
				break;
			}
			case selection: {
				select(table);
				Dispatch selection = getSelection(app);

				Dispatch cells = Dispatch.get(selection, "Cells").getDispatch();
				EnumVariant enumVariant = new EnumVariant(cells);

				for (int i = 0; i < rowCount; ++i) {
					for (int j = 0; j < colCount; ++j) {
						if (enumVariant.hasMoreElements()) {
							Dispatch cell = enumVariant.nextElement().toDispatch();
							String value = values[i][j];
							if (!value.isEmpty()) {
								createRangeFromRangeOf(cell).setText(value);
							}
						}
					}
				}
				break;
			}
			case cellIndexing: {
				for (int i = 0; i < rowCount; ++i) {
					int wordRowIdx = i + 1; // vba starts with 1
					for (int j = 0; j < colCount; ++j) {
						int wordColumnIdx = j + 1; // vba starts with 1
						Dispatch cell = Dispatch.call(table, "Cell",
								DocWordHelper.getVariant(wordRowIdx),
								DocWordHelper.getVariant(wordColumnIdx)).toDispatch();
						Range<Dispatch> cellRange = createRangeFromRangeOf(cell);
						cellRange.setText(values[i][j]);
					}
				}
				break;
			}
		}
	}

	private static final Variant wdLineStyleSingle = DocWordHelper.getVariant(1);
	private static final Variant wdPreferredWidthPercent = DocWordHelper.getVariant(2);
	private static final Variant wdPreferredWidthPoints = DocWordHelper.getVariant(3);
	private static final Variant wdTexture10Percent = DocWordHelper.getVariant(100);
	private static final Variant wdTexture25Percent = DocWordHelper.getVariant(250);

	private void formatTable(Dispatch table, PropertiesDoc doc, Style tabhead, Level level) {
		long start = System.currentTimeMillis();
		Dispatch.put(table, "PreferredWidthType", wdPreferredWidthPercent);
		Dispatch.put(table, "PreferredWidth", DocWordHelper.getVariant(100));
		long pause = System.currentTimeMillis();
		_logger.log(level, "    ... " + (pause - start) + " ms for widths (table)");

		start = pause;
		EnumVariant columns = getIterator(table, "Columns");
		for (ColumnSpec colSpec : doc.getTableSpec().getColSpecs()) {
			Dispatch column = columns.nextElement().getDispatch();
			Dispatch.put(column, "PreferredWidthType", wdPreferredWidthPoints);
			Dispatch.put(column, "Width", new Variant(pointForPerc(colSpec.getRelWidth())));
		}
		pause = System.currentTimeMillis();
		_logger.log(level, "    ... " + (pause - start) + " ms for widths (columns)");

		start = pause;
		// EnumVariant rows = getIterator(table, "Rows");
		// For the sake of optimisation, we don't use iterator here, but the direct access to
		// an item with index, because we need to access rows only in case of special formatting:
		Dispatch rows = Dispatch.call(table, "Rows").toDispatch();
		for (int i = 0; i < doc.getRowCount(); ++i) {
			Kind kind = doc.getEntryDocs().get(i).getKind();
			// for "pure" data, we don't need formatting, created table already uses cell style
			if (kind != EntryDoc.Kind.data) {
				Dispatch row = DocWordHelper.getItem(rows, i + 1);
				if (kind == EntryDoc.Kind.columnLabels) {
					Dispatch.put(row, "HeadingFormat", DocWordHelper.VARIANT_TRUE);
				} else if (kind == EntryDoc.Kind.tableName) {
					Dispatch.put(row, "HeadingFormat", DocWordHelper.VARIANT_TRUE);
					mergeCells(row);
					applyShaddingToRow(row, wdTexture25Percent);
				} else if (kind == EntryDoc.Kind.groupSubhead) {
					mergeCells(row);
					applyShaddingToRow(row, wdTexture10Percent);
				}
				setRangeStyle(createRangeFromRangeOf(row).getObject(), tabhead);
			}
		}
		pause = System.currentTimeMillis();
		_logger.log(level, "    ... " + (pause - start) + " ms for shadding/merging/styling (rows)");

		start = pause;
		Dispatch borders = Dispatch.get(table, "Borders").toDispatch();
		Dispatch.put(borders, "InsideLineStyle", wdLineStyleSingle);
		Dispatch.put(borders, "OutsideLineStyle", wdLineStyleSingle);
		pause = System.currentTimeMillis();
		_logger.log(level, "    ... " + (pause - start) + " ms for borders");
	}

	private static void mergeCells(Dispatch row) {
		Dispatch cells = Dispatch.get(row, "Cells").toDispatch();
		Dispatch.call(cells, "Merge");
	}

	private static void applyShaddingToRow(Dispatch row, Variant shadingLevel) {
		Dispatch shading = Dispatch.call(row, "Shading").toDispatch();
		Dispatch.put(shading, "Texture", shadingLevel);
	}

	// FIXME table HTML; should be private when done
	void fillCellWithHtml(Dispatch table, int i, int j, boolean isFormatted, String val, Style style) {
		Dispatch cell = Dispatch.call(table, "Cell", DocWordHelper.getVariant(i),
				DocWordHelper.getVariant(j)).toDispatch();

		Range<Dispatch> cellRange = createRangeFromRangeOf(cell);
		if (isFormatted) {
			cellRange.setEnd(cellRange.getEnd() + 1);
			setRangeStyle(cellRange.getObject(), style);
			// it's faster to check whether empty, then to clear clipboard:
			if (!val.trim().isEmpty()) {
				backspaceLast(cellRange); // eat EOL added above
				Util.copyHtmlToClipboard(val);
				try {
					pasteHtmlFromClipboard(cellRange.getObject());
				} catch (ApplicationException e) {
					_logger.warn(Arrays.deepToString(e.getStackTrace()));
				}
				backspaceLast(cellRange); // eat EOL - paste HTML from clipboard adds EOL
			}
			setRangeStyle(cellRange.getObject(), style);
		} else {
			setRangeStyle(cellRange.getObject(), style);
			cellRange.setText(val);
		}
	}

	private void backspaceLast(Range<Dispatch> range) {
		int oneBeforeEnd = range.getEnd() - 1;
		range.setStart(oneBeforeEnd);
		range.setEnd(oneBeforeEnd);
		Dispatch.call(range.getObject(), "delete");
	}

	// ------------------------------ for debugging only ---------------------------

	// private Range<Dispatch> getField(Dispatch scope, int idx) {
	// Dispatch fields = Dispatch.get(scope, "Fields").toDispatch();
	// int sz = DocWordHelper.getCount(fields);
	// if (sz < idx) {
	// return null;
	// }
	// Dispatch field = DocWordHelper.getItem(fields, idx);
	// return createRange(Dispatch.get(field, "Code").toDispatch());
	// }
	//
	// private String printFields(Dispatch scope) {
	// StringBuilder sb = new StringBuilder();
	// Dispatch fields = Dispatch.get(scope, "Fields").toDispatch();
	// int sz = DocWordHelper.getCount(fields);
	// for (int i = 1; i <= sz; ++i) {
	// Dispatch field = DocWordHelper.getItem(fields, i);
	// Range<Dispatch> code = createRange(Dispatch.get(field, "Code").toDispatch());
	// sb.append("'").append(code.getText()).append("'").append(Util.NL);
	// }
	// return sb.toString();
	// }

	// void addParaForTable(Range<Dispatch> range) {
	// appendNewLine(range); // does NOT create new paragraph!
	// collapseRangeToEnd(range);
	// prependNewLine(range); // this one DOES
	// collapseRangeToStart(range);
	// }

	// int getDocumentParagraphCount(Dispatch document) {
	// Dispatch paragraphs = Dispatch.get(document, "Paragraphs").toDispatch();
	// return getCount(paragraphs);
	// }

	// static void save(Dispatch doc, String outputFileAbsPath) {
	// _logger.info("... saving MSWord file '" + outputFileAbsPath + "'");
	// Dispatch.call(doc, "Save");
	// }

	// ----------------- document lifecycle -----------------

	@Override
	public void createWordApp() {
		_logger.info("creating MSWord application");
		ActiveXComponent ax = new ActiveXComponent("Word.Application");
		boolean isVisible = false;
		ax.setProperty("Visible", DocWordHelper.getVariant(isVisible));
		_wordApp = ax.getObject();
	}

	@Override
	public String getWordAppName() {
		return Dispatch.get(_wordApp, "Name").getString();
	}

	@Override
	public String getWordAppVersion() {
		return Dispatch.get(_wordApp, "Version").getString();
	}

	@Override
	public void openDoc() {
		_logger.info("  opening MSWord file '" + getInput().getOutFilePath() + "'");
		Dispatch documents = Dispatch.get(_wordApp, "Documents").toDispatch();
		_wordDoc = Dispatch.call(documents, "Open", getInput().getOutFilePath()).toDispatch();

		// this is somewhat faster (initialise once per opening):
		_bookmarks = Dispatch.call(_wordDoc, "Bookmarks").toDispatch();
		_hyperlinks = Dispatch.call(_wordDoc, "Hyperlinks").toDispatch();
	}

	@Override
	public void closeDoc() {
		if (_wordDoc != null) {
			_logger.info("closing MSWord file '" + getInput().getOutFilePath() + "'");
			boolean saveChanges = true;
			Dispatch.call(_wordDoc, "Close", DocWordHelper.getVariant(saveChanges));
		}
	}

	@Override
	public void exitAppAndSaveDocument() {
		_logger.info("exiting MSWord application");
		boolean saveOnExit = true;
		Dispatch.call(_wordApp, "Quit", DocWordHelper.getVariant(saveOnExit));
	}

	// ------------------- application & document optimisation options ---------------

	private OptimOptions _initialOptions;
	private OptimOptions _optimisedOptions;

	@Override
	public void initDocgenOptimisationOptions() {
		if (_wordDoc == null) {
			return;
		}

		_initialOptions = new OptimOptions(_wordDoc, _wordApp);
		_logger.debug("original(_initialOptions): " + _initialOptions.toString());
	}

	@Override
	public void setDocgenOptimisationOptions() {
		if (_wordDoc == null) {
			return;
		}

		// copy original state and change if necessary:
		_optimisedOptions = new OptimOptions(_initialOptions);
		_optimisedOptions.setForOptimisedExecution(_wordDoc, _wordApp);
		_logger.debug("current(_optimisedOptions): " + _optimisedOptions.toString());
	}

	@Override
	public void unsetDocgenOptimisationOptions() {
		if (_wordDoc == null) {
			return;
		}

		_logger.debug("  restoring original options");
		_optimisedOptions.restoreFrom(_initialOptions, _wordDoc, _wordApp);
		_logger.debug(">>  current/restored          : " + _optimisedOptions.toString());

		_optimisedOptions = null; // not used after this point
	}

	// --------- bookmarks and hyperlinks -------------

	@Override
	public void insertBookmark(Range<Dispatch> range, String label) {
		Dispatch.call(_bookmarks, "Add", label, range.getObject());
	}

	@Override
	public void insertHyperlink(Range<Dispatch> range, String textToDisplay, String url) {
		Dispatch.call(_hyperlinks, "Add", range.getObject(), url, textToDisplay);
	}

	// --------- styles -------------

	@Override
	public Map<String, String> getExistingStyleNames() {
		_logger.info("   collecting existing style names....");
		long start = System.currentTimeMillis();

		Map<String, String> names = new HashMap<String, String>();
		EnumVariant styles = getIterator(_wordDoc, "Styles");
		while (styles.hasMoreElements()) {
			Dispatch style = styles.nextElement().getDispatch();
			String name = Dispatch.get(style, "NameLocal").getString();
			names.put(name, name);
		}
		Util.logCompletion(Level.INFO, "   collected existing style names", start, getInput()
				.isSkipTiming());
		return names;
	}

	// ---------------- TOCs ---------------

	@Override
	public void updateFields() {
		long start = System.currentTimeMillis();

		Dispatch.call(DocWordHelper.getFields(_wordDoc), "Update");

		long pause = System.currentTimeMillis();
		_logger.info("   updateFields: " + (pause - start) + " ms");
	}

	@Override
	public void updateTablesOf(String what) {
		long start = System.currentTimeMillis();

		EnumVariant tocs = getIterator(_wordDoc, what);
		while (tocs.hasMoreElements()) {
			Dispatch toc = tocs.nextElement().getDispatch();
			Dispatch.call(toc, "Update");
		}

		long pause = System.currentTimeMillis();
		_logger.info("     update" + what + ": " + (pause - start) + " ms");
	}

	// ----------------------- custom properties -----------------------

	// string as type for custom property:
	private static final Variant msoPropertyTypeString = DocWordHelper.getVariant(4);

	@Override
	public Map<String, String> getCustomDocProperties() {
		Map<String, String> result = new HashMap<String, String>();
		EnumVariant customProps = getIterator(_wordDoc, "CustomDocumentProperties");
		while (customProps.hasMoreElements()) {
			Dispatch customProp = customProps.nextElement().toDispatch();
			String name = Dispatch.get(customProp, "Name").getString();
			String value = Dispatch.get(customProp, "Value").getString();
			result.put(name, value);
		}
		return result;
	}

	@Override
	public void setCustomDocProperties(Map<String, String> newCustomProps) {
		Map<String, String> newProps = new HashMap<String, String>(newCustomProps);
		Dispatch customProps = Dispatch.get(_wordDoc, "CustomDocumentProperties").toDispatch();
		int count = DocWordHelper.getCount(customProps);
		for (int i = 1; i <= count; ++i) {
			Dispatch customProp = DocWordHelper.getItem(customProps, i);
			String name = Dispatch.get(customProp, "Name").getString();
			if (newProps.containsKey(name)) {
				// "move" from requested map into Word document:
				String value = newProps.remove(name);
				_logger.debug(String.format("setting existing custom property: %s = %s ", name,
						value));
				Dispatch.put(customProp, "Value", new Variant(value));
			}
		}
		if (newProps.isEmpty()) {
			// all requested properties existed in the Word doc and their new values have been set
			return;
		}

		for (Map.Entry<String, String> newProp : newProps.entrySet()) {
			String name = newProp.getKey();
			String value = newProp.getValue();
			_logger.info(String.format("adding new custom property: %s = %s ", name, value));

			Dispatch.call(customProps, "Add", new Variant(name), DocWordHelper.getVariant(false),
					msoPropertyTypeString, new Variant(value));
		}
	}

	// --------------------- cursors ---------------------

	/**
	 * {@inheritDoc}
	 * <p>
	 * Implementation of closing/reopening the MS Word document.
	 */
	@Override
	public Cursor<Dispatch> closeAndReopenDoc(CursorList<Dispatch> cursors,
			Cursor<Dispatch> currentCursor) {
		long t0 = System.currentTimeMillis();

		List<Map.Entry<Integer, Integer>> indexes = cursors.snapshotIndexes();
		closeDoc();

		openDoc();
		setDocgenOptimisationOptions();

		Range<Dispatch> range = getDocumentAsRange();

		Cursor<Dispatch> result = null;
		for (int i = 0; i < cursors.size(); ++i) {
			Cursor<Dispatch> oldCursor = cursors.get(i);
			Placeholder ph = oldCursor.getPlaceholder();
			Map.Entry<Integer, Integer> startAndEnd = indexes.get(i);
			int start = startAndEnd.getKey().intValue();
			int end = startAndEnd.getValue().intValue();

			Range<Dispatch> limited = duplicateRange(range);
			limited.setStartEnd(start, end);

			Cursor<Dispatch> newCursor = createCursor(ph, limited);
			cursors.set(i, newCursor);
			if (ph == currentCursor.getPlaceholder()) {
				result = newCursor;
			}
		}
		Util.logCompletion(Level.INFO, "closed and reopened document.", t0, getInput()
				.isSkipTiming());

		return result;
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * We have tried with GetCrossReferenceItems on document, but that was not reliable.
	 */
	@Override
	public Map<CaptionKind, List<Caption<Dispatch>>> collectCaptions() {
		Map<CaptionKind, List<Caption<Dispatch>>> result = new LinkedHashMap<CaptionKind, List<Caption<Dispatch>>>();
		List<Caption<Dispatch>> figures = new ArrayList<Caption<Dispatch>>();
		result.put(CaptionKind.Figure, figures);
		List<Caption<Dispatch>> tables = new ArrayList<Caption<Dispatch>>();
		result.put(CaptionKind.Table, tables);

		EnumVariant paras = getIterator(getDocumentContent(), "Paragraphs");
		while (paras.hasMoreElements()) {
			Dispatch para = paras.nextElement().getDispatch();

			String styleName = getParagraphStyleName(para);
			Range<Dispatch> range = createRangeFromRangeOf(para);
			String text = range.getText();

			if (CaptionKind.looksLikeFigureCaption(styleName, text)) {
				Caption<Dispatch> caption = createCaption(CaptionKind.Figure, range);
				_logger.info("   found '" + text + "'");
				figures.add(caption);
			} else if (CaptionKind.looksLikeTableCaption(styleName, text)) {
				Caption<Dispatch> caption = createCaption(CaptionKind.Table, range);
				_logger.info("   found '" + text + "'");
				tables.add(caption);
			}
		}

		return result;
	}

	@Override
	public Range<Dispatch> createRange(Dispatch object) {
		return new DocWordRange(object);
	}

	@Override
	public final WordPatternFinder<Dispatch> createPatternFinder(String msPattern) {
		return new DocWordPatternFinder(_wordDoc, msPattern);
	}

	// ---------------- actual writing into output document -------------------

	@Override
	public void clearUndoCache() {
		Dispatch.call(_wordDoc, "UndoClear");
	}

	@Override
	public Range<Dispatch> getDocumentAsRange() {
		return createRange(getDocumentContent());
	}

	@Override
	public Range<Dispatch> duplicateRange(Range<Dispatch> range) {
		return createRange(Dispatch.call(range.getObject(), "Duplicate").toDispatch());
	}

	@Override
	public void prependText(Range<Dispatch> range, String newText) {
		Dispatch.call(range.getObject(), "InsertBefore", newText);
	}

	@Override
	public int getRangeParagraphOutlineLevel(Range<Dispatch> range, int paraIdx) {
		Dispatch paragraph = getParagraph(range.getObject(), paraIdx);
		if (paragraph == null) {
			return -1;
		}
		return Dispatch.get(paragraph, "OutlineLevel").getInt();
	}

	@Override
	public String appendRawTextInNewParagraphWithStyle(Range<Dispatch> range, String newText,
			Style style) {
		appendNewLine(range);
		return appendTextWithStyle(range, newText, style);
	}

	@Override
	public String appendHtmlTextInNewParagraphWithStyle(Range<Dispatch> range, String newMarkup,
			Style style) {
		System.out.println(">>> " + range.getText());
		System.out.println("    " + newMarkup);
		if (newMarkup.trim().isEmpty()) {
			return "";
		}
		appendNewLine(range);
		collapseRangeToEnd(range);
		range.setEnd(range.getEnd() + 1);
		setRangeStyle(range.getObject(), style);
		Util.copyHtmlToClipboard(newMarkup);
		try {
			pasteHtmlFromClipboard(range.getObject());
		} catch (ApplicationException e) {
			_logger.warn(Arrays.deepToString(e.getStackTrace()));
		}
		// backspaceLast(range); // eat EOL - pasting HTML from clipboard adds an EOL
		return range.getText();
	}

	@Override
	public String appendTextWithStyle(Range<Dispatch> range, String newText, Style style) {
		collapseRangeToEnd(range);
		setRangeStyle(range.getObject(), style);
		return appendText(range, newText);
	}

	// ------ inserting a figure/table, its caption, and references to caption ------

	private static final Variant wdCaptionFigure = DocWordHelper.getVariant(-1);
	private static final Variant wdCaptionTable = DocWordHelper.getVariant(-2);
	private static final Variant wdCaptionPositionAbove = DocWordHelper.getVariant(0);
	private static final Variant wdCaptionPositionBelow = DocWordHelper.getVariant(1);
	private static final Variant wdParagraph = DocWordHelper.getVariant(4);
	private static final Variant wdOnlyLabelAndNumber = DocWordHelper.getVariant(3);

	@Override
	public void insertFigureCaption(Range<Dispatch> range, String text, String logMsg) {
		insertCaption(range, wdCaptionFigure, wdCaptionPositionBelow, text, logMsg);

		int paraCount = 2;
		Dispatch.call(range.getObject(), "MoveEnd", wdParagraph,
				DocWordHelper.getVariant(paraCount));

		Dispatch secondPara = getParagraph(range.getObject(), paraCount);
		setRangeStyle(secondPara, Style.figcapt);

		int endMinusOne = range.getEnd() - 1;
		range.setEnd(endMinusOne);
		collapseRangeToEnd(range);
	}

	@Override
	public void insertTableCaption(Range<Dispatch> range, String text, String logMsg) {
		Range<Dispatch> tableRange = getFirstTableAsRange(range.getObject());

		insertCaption(tableRange, wdCaptionTable, wdCaptionPositionAbove, text, logMsg);

		Dispatch firstPara = getParagraph(tableRange.getObject(), 1);
		setRangeStyle(firstPara, Style.tabcapt);

		collapseRangeToEnd(range);

		firstPara = getParagraph(range.getObject(), 1);
		setRangeStyle(firstPara, Style.para); // was tabcell, set by insertTable, using keepWithNext
	}

	private static void insertCaption(Range<Dispatch> range, Variant label, Variant position,
			String text, String logMsg) {
		long t0 = System.currentTimeMillis();

		String textAfterCaptionNumber = " " + Util.EN_DASH + " " + text;

		Dispatch.call(range.getObject(), "InsertCaption", label,
				new Variant(textAfterCaptionNumber), DocWordHelper.VARIANT_EMPTY_STRING, position,
				DocWordHelper.getVariant(0));

		_logger.info("... " + (System.currentTimeMillis() - t0) + " ms for insertCaption(): "
				+ logMsg + textAfterCaptionNumber);
	}

	@Override
	public void insertFigureRef(Range<Dispatch> insertPointRange, int figNumber) {
		insertCaptionRef(insertPointRange, wdCaptionFigure, figNumber);
	}

	@Override
	public void insertTableRef(Range<Dispatch> insertPointRange, int tabNumber) {
		insertCaptionRef(insertPointRange, wdCaptionTable, tabNumber);
	}

	private void insertCaptionRef(Range<Dispatch> range, Variant refType, int idx) {
		long t0 = System.currentTimeMillis();

		prependText(range, " ");
		collapseRangeToStart(range);

		// InsertCrossReference(ReferenceType, ReferenceKind, ReferenceItem,
		// InsertAsHyperlink, IncludePosition, SeparateNumbers, SeparatorString)
		Dispatch.call(range.getObject(), "InsertCrossReference", refType, wdOnlyLabelAndNumber,
				DocWordHelper.getVariant(idx), DocWordHelper.getVariant(true),
				DocWordHelper.getVariant(false), DocWordHelper.getVariant(false),
				DocWordHelper.VARIANT_EMPTY_STRING);
		_logger.debug("... " + (System.currentTimeMillis() - t0) + " ms for insertCaptionRef()");
		_logger.info("--------------------------------------------------");
	}

	@Override
	public void insertFigure(Range<Dispatch> range, File pic) {
		range.setText("");
		// "Link to File" = false, "Save with Document" = true"
		Dispatch inlineShapes = Dispatch.call(range.getObject(), "InlineShapes").toDispatch();
		Dispatch.call(inlineShapes, "AddPicture", pic.getAbsolutePath(),
				DocWordHelper.VARIANT_FALSE, DocWordHelper.VARIANT_TRUE);
		range.setEnd(range.getEnd() + 1);
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Returns expanded range to the end of the table. We do not collapse original range here, as we
	 * assume the caption will need to be inserted before the table.
	 */
	@Override
	public int insertTable(Range<Dispatch> range, PropertiesDoc doc, Style tabhead,
			boolean addBookmarks) {
		return doInsertTable(range, doc, tabhead, addBookmarks);
	}

	// ------------------------------ tests only ---------------------------

	@Override
	public String getRangeParagraphStyleName(Range<Dispatch> range, int paraIdx) {
		Dispatch paragraph = getParagraph(range.getObject(), paraIdx);
		return (paragraph == null) ? null : getParagraphStyleName(paragraph);
	}

	private Dispatch getParagraph(Dispatch range, int paraIdx) {
		Dispatch paragraphs = Dispatch.get(range, "Paragraphs").toDispatch();
		int count = DocWordHelper.getCount(paragraphs);
		if (paraIdx < 1 || paraIdx > count) {
			return null;
		}
		return DocWordHelper.getItem(paragraphs, paraIdx);
	}

	private String getParagraphStyleName(Dispatch paragraph) {
		Dispatch style = Dispatch.get(paragraph, "Style").toDispatch();
		return Dispatch.get(style, "NameLocal").getString();
	}

	@Override
	public int getRangeParagraphCount(Range<Dispatch> range) {
		Dispatch paragraphs = Dispatch.get(range.getObject(), "Paragraphs").toDispatch();
		return DocWordHelper.getCount(paragraphs);
	}

	@Override
	public void prependNewLine(Range<Dispatch> range) {
		Dispatch.call(range.getObject(), "InsertParagraphBefore");
	}

	@Override
	public void appendNewLine(Range<Dispatch> range) {
		Dispatch.call(range.getObject(), "InsertParagraphAfter");
	}

	@Override
	public String appendText(Range<Dispatch> range, String newText) {
		Dispatch.call(range.getObject(), "InsertAfter", newText);
		return range.getText();
	}

	@Override
	public String appendTextInNewParagraph(Range<Dispatch> range, String newText) {
		appendNewLine(range);
		Dispatch.call(range.getObject(), "InsertAfter", newText);
		return range.getText();
	}

	@Override
	public boolean isRangeWithTable(Dispatch range) {
		return null != getFirstTableAsRange(range);
	}

	private Range<Dispatch> getFirstTableAsRange(Dispatch container) {
		Dispatch tables = Dispatch.get(container, "Tables").toDispatch();
		Dispatch table = DocWordHelper.getItem(tables, 1);
		return createRangeFromRangeOf(table);
	}

	private static final Variant wdCollapseStart = DocWordHelper.getVariant(1);
	private static final Variant wdCollapseEnd = DocWordHelper.getVariant(0);

	@Override
	public void collapseRangeToEnd(Range<Dispatch> range) {
		Dispatch.call(range.getObject(), "Collapse", wdCollapseEnd);
	}

	private void collapseRangeToStart(Range<Dispatch> range) {
		Dispatch.call(range.getObject(), "Collapse", wdCollapseStart);
	}

	// ===== Impl. of org.tanjakostic.jcleancim.docgen.writer.word.WordWriter methods =====

	@Override
	public boolean applyCloseReopen() {
		return getInput().getSaveRecloseEvery() != -1;
	}

	// ===== Impl. of org.tanjakostic.jcleancim.docgen.writer.Writer methods =====

	public static final List<String> FILE_EXTENSIONS = Arrays.asList(".doc", ".docx");

	@Override
	public Set<String> getSupportedFormats() {
		return new HashSet<String>(FILE_EXTENSIONS);
	}
}
