/**
 * Copyright (C) 2009-2016 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.model;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.tanjakostic.jcleancim.util.Util;

/**
 * UML stereotype.
 * <p>
 * It is essentially a set of comma-separated string values. To avoid the application to do string
 * parsing and adding e.g. "<<" and ">>" around the stereotypes, this simple class does it in one
 * place.
 * <p>
 * FIXME: At present, we don't check for duplicates, and no modifiers...
 *
 * @author tatjana.kostic@ieee.org
 * @version $Id: UmlStereotype.java 15 2016-07-12 15:11:42Z dev978 $
 */
public class UmlStereotype {

	/** UML stereotype for enumeration class. */
	public static final String ENUMERATION = "enumeration";

	/** UML stereotype for enumeration literals. */
	public static final String ENUM = "enum";

	/** UML stereotype for deprecated UML elements or connectors. */
	public static final String DEPRECATED = "deprecated";

	/** UML stereotype for non-normative associations. */
	public static final String INFORMATIVE = "informative";

	// CIM class stereotypes; 'enumeration' is UML standard, others are ours.
	public static final String PRIMITIVE = "Primitive";
	public static final String OLD_DATATYPE = "Datatype"; // for backwards compatibility
	public static final String CIMDATATYPE = "CIMDatatype";
	public static final String COMPOUND = "Compound";

	public static final List<String> CIM_STEREOTYPES = Arrays.asList(ENUMERATION, ENUM, DEPRECATED,
			INFORMATIVE, PRIMITIVE, OLD_DATATYPE, CIMDATATYPE, COMPOUND);

	// WG10 class stereotypes. 'interface' is UML standard, others are ours
	public static final String EA_INTERFACE = "interface";
	public static final String PACKED = "packed";
	public static final String BASIC = "basic";
	public static final String STRUCTURED = "structured";

	/**
	 * Used for some abstract LN classes (61850-7-4) to tag that presence conditions of its data
	 * objects do not change in the context of derived statistic instance. Not inheritable (i.e.,
	 * should not be printed in subclasses).
	 **/
	public static final String ADMIN = "admin";

	/**
	 * Used for CDCs (61850-7-3) alowed for use as type in DOs of derived statistics LNs; CDCs
	 * without this stereotype are forbidden for use in derived statistics context. It is
	 * inheritable (i.e., should be printed in subclasses).
	 **/
	public static final String STATISTICS = "statistics";

	/** Used for enumerations that represent presence conditions (constraints). */
	public static final String COND = "cond";

	/** Used for enumerations that represent abbreviations. */
	public static final String ABBR = "abbr";

	/** Used for operations. */
	public static final String EVENT = "event";

	public static final List<String> IEC61850_STEREOTYPES = Arrays.asList(ENUMERATION, ENUM,
			DEPRECATED, INFORMATIVE, EA_INTERFACE, PACKED, BASIC, STRUCTURED, ADMIN, STATISTICS,
			COND, ABBR, EVENT);

	// -------------------

	private final Set<String> _tokens = new LinkedHashSet<String>();
	private String _commaSeparatedTokens; // cache

	/**
	 * Constructor.
	 *
	 * @param tokens
	 *            (optional) desired number of individual non-null stereotype tokens; empty tokens
	 *            are skipped. A token that contains comma-separated items will be split and each of
	 *            those items will be kept as a stereotype token.
	 */
	public UmlStereotype(String... tokens) {
		Util.ensureContainsNoNull(tokens, "tokens");
		for (String token : tokens) {
			List<String> splitSubtokens = Util.splitCommaSeparatedTokens(token);
			_tokens.addAll(splitSubtokens);
		}
	}

	private String buildCommaSeparatedToken() {
		return Util.concatCharSeparatedTokens(", ", false, new ArrayList<String>(_tokens));
	}

	/** Returns whether this instance is empty (has no tokens). */
	public boolean isEmpty() {
		return _tokens.isEmpty();
	}

	/** Returns whether this instance contains <code>token</code>). */
	public boolean contains(String token) {
		return _tokens.contains(token);
	}

	/** Returns whether all the tokens of this instance are contained in <code>tokens</code>. */
	public boolean memberOf(Set<String> tokens) {
		return getMissingTokens(tokens).isEmpty();
	}

	/** Returns all the tokens of this instance that are not contained in <code>tokens</code>. */
	public Set<String> getMissingTokens(Set<String> tokens) {
		Set<String> result = new LinkedHashSet<String>(_tokens);
		result.removeAll(tokens);
		return result;
	}

	/** Returns comma-separated list of stereotype tokens. */
	public String value() {
		if (_commaSeparatedTokens == null) {
			_commaSeparatedTokens = buildCommaSeparatedToken();
		}
		return _commaSeparatedTokens.toString();
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Ensloses {@link #value()} into "<<" and ">>".
	 */
	@Override
	public String toString() {
		return (_tokens.isEmpty()) ? "" : ("<<" + value() + ">>");
	}
}
